<?php

namespace FiloBlu\Flow\Model\Channel\In;

use Magento\ImportExport\Model\Import\Adapter as ImportAdapter;
use Magento\Framework\App\Filesystem\DirectoryList;
use Magento\ImportExport\Model\Import;

class Images extends AbstractModel {

    protected $_logger;
    protected $productFactory;
    protected $directoryList;
    protected $scopeConfig;
    protected $imagesHelper;

    private $channel_photo_config;
    private $collection_of_images_to_parse;
    private $array_of_data_for_csv;
    private $filename_separator = "_";
    private $image_alt_separator = " ";
    private $eventManager;
    private $errors_processing_images = 0;

    private $cache_folder = BP . "/pub/media/catalog/product/cache/";

    public function __construct(
        \FiloBlu\Flow\Helper\LoggerProvider $loggerProvider,
        \FiloBlu\Flow\Model\Channel\Config $channelConfig,
        \Magento\Catalog\Model\ProductFactory $productFactory,
        \Magento\Framework\App\Filesystem\DirectoryList $directoryList,
        \FiloBlu\Flow\Helper\Images $imagesHelper,
        \Magento\Framework\App\Config\ScopeConfigInterface $scopeConfig,
        \Magento\Framework\Event\Manager $eventManager
    ) {
        parent::__construct($loggerProvider, $channelConfig);
        $this->_logger = $loggerProvider->getLogger();
        $this->productFactory = $productFactory;
        $this->directoryList = $directoryList;
        $this->scopeConfig = $scopeConfig;
        $this->imagesHelper = $imagesHelper;
        $this->eventManager = $eventManager;
    }

    public function load($channelname) {
        $this->init($channelname);
    }

    public function insertData($file){
        $meta_file_id = $file->getId();

        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\Imagesflow');

        // Loading images SKUs to process
        $collection = $model->getCollection();
        $collection->addFieldToSelect(array('id','image_sku'));
        $collection->addFieldToFilter("meta_processed",0);
        $collection->addFieldToFilter("meta_file",$meta_file_id);
        $collection->addFieldToFilter("image_status","parsed");
        $collection->getSelect()->group('image_sku');
        $collection->load();

        $iterator = \Magento\Framework\App\ObjectManager::getInstance()->create('Magento\Framework\Model\ResourceModel\Iterator');
        $iterator->walk($collection->getSelect(), array(array($this, 'imagesCallback')),array('meta_file_id' => $meta_file_id));

        if($this->errors_processing_images > 0){
            return false;
        } else {
            return true;
        }

    }


    public function imagesCallback($args){

        $objectManager = \Magento\Framework\App\ObjectManager::getInstance(); // Instance of object manager

        $sku = $args['row']['image_sku'];
        $meta_file_id = $args['meta_file_id'];

        $product = $this->productFactory->create();
        $product_id = $product->getIdBySku($sku);

        if (!$product_id) {
            $this->_logger->info("INSERT IMAGES: {$sku} not found in Magento");
            $this->errors_processing_images = $this->errors_processing_images + 1;
            return false;
        }

        $this->_logger->info("INSERT IMAGES: Running images import for SKU {$sku}");

        $model = $objectManager->create('FiloBlu\Flow\Model\Imagesflow');
        //$product = $objectManager->create('Magento\Catalog\Model\Product')->setStoreId(0)->load($product_id);
        $product = $this->productFactory->create();
        $product->setStoreId(0)->load($product_id);

        // Loading images SKUs to process
        $collection = $model->getCollection();
        $collection->addFieldToFilter("meta_processed",0);
        $collection->addFieldToFilter("meta_file",$meta_file_id);
        $collection->addFieldToFilter("image_status","parsed");
        $collection->addFieldToFilter("image_sku",$sku);
        $collection->setOrder('image_position', 'ASC');
        $collection->load();

        // Array for later use
        $position_to_filename_array = array();
        foreach ($collection as $image_data) {
            $position_to_filename_array[$image_data->getImagePosition()]['image_name'] = $image_data->getImageName();
            $position_to_filename_array[$image_data->getImagePosition()]['image_name_no_ext'] = $this->removeExtensionFromFilename($image_data->getImageName());
            $position_to_filename_array[$image_data->getImagePosition()]['image_alt'] = $image_data->getImageAlt();
        }

        //Remove the images on the positions where we will put the new ones
        //$productRepository = $objectManager->create('Magento\Catalog\Api\ProductRepositoryInterface');
        $existingMediaGalleryEntries = $product->getMediaGalleryEntries();
        $this->_logger->info("INSERT IMAGES: Removing images before replace them from SKU {$sku}");
        $errors_on_removing = 0;
        if (count($existingMediaGalleryEntries)) {
            foreach ($existingMediaGalleryEntries as $key => $entry) {
                if (isset($position_to_filename_array[$entry->getPosition()]) || $this->imagesHelper->isFullPhotoCleanActive()) {
                    //unset($existingMediaGalleryEntries[$key]);
                    $done = $this->removeCatalogProductImageByPath($entry->getFile());
                    if(!$done){
                        $errors_on_removing++;
                    }
                }
            }
            //$product->setMediaGalleryEntries($existingMediaGalleryEntries);
            //$productRepository->save($product);
        }

        if($errors_on_removing){
            $this->_logger->info("INSERT IMAGES: Error removing images from {$sku}, can't go ahead!");
            $this->errors_processing_images = $this->errors_processing_images + 1;
            return false;
        }

        $this->_logger->info("INSERT IMAGES: Removing images for SKU {$sku} done!");

        $product = $this->productFactory->create();
        $product->setStoreId(0)->load($product_id);

        $this->_logger->info("INSERT IMAGES: Start inserting new images for SKU {$sku}");
        //Now let's insert the new images
        $photos_added = 0;
        foreach ($collection as $image_data) {

            $importDir = $this->directoryList->getPath('media').DIRECTORY_SEPARATOR.'import'.DIRECTORY_SEPARATOR;

            if (!file_exists($importDir)) {
                mkdir($importDir, 0777, true);
            }

            //Move filename under pub media because Magento reads from there
            //TODO: Change this -> setting Magento to read directly from var/bouncer
            $imagePath = $this->directoryList->getPath('var').DIRECTORY_SEPARATOR.$image_data->getImagePath().DIRECTORY_SEPARATOR.$image_data->getImageName(); // path of the image
            $newImagePath = $importDir.$image_data->getImageName();
            $array_image_type = explode("|", $image_data->getImageType());

            if (!file_exists($imagePath)) {
                $image_data->setData('meta_processed','1');
                $image_data->setLog(trim("IMAGE: {$imagePath} not found\n".$image_data->getLog()));
                $image_data->save();
                $this->_logger->info("IMAGE: {$imagePath} not found");
                $this->errors_processing_images = $this->errors_processing_images + 1;
                continue;
            }

            $pathinfo = pathinfo($image_data->getImageName());
            $imgExtensions = ['jpg', 'jpeg', 'gif', 'png'];
            if (!isset($pathinfo['extension']) || !in_array(strtolower($pathinfo['extension']), $imgExtensions)) {
                $image_data->setData('meta_processed','1');
                $image_data->setLog(trim("FILENAME is incorrect: {$imagePath}\n".$image_data->getLog()));
                $image_data->save();
                $this->errors_processing_images = $this->errors_processing_images + 1;
                continue;
            }

            copy($imagePath, $newImagePath);

            $product->addImageToMediaGallery($newImagePath, $array_image_type, true, $image_data->getImageHide());
            $photos_added++;
            $this->_logger->info("INSERT IMAGES: Added {$imagePath} on SKU {$sku}");
        }

        if (!$photos_added){
            $this->_logger->info("INSERT IMAGES: No new photos added on SKU {$sku}");
            return false;
        }

        $this->_logger->info("INSERT IMAGES: Saving product {$sku}");
        try {
            $product->save();
        } catch (\Exception $e) {
            $this->_logger->info("ERROR IMAGES: Saving product {$sku} with message " . $e->getMessage());
            $this->errors_processing_images = $this->errors_processing_images + 1;
            return false;
        }
        $this->_logger->info("INSERT IMAGES: Saved product {$sku}");

        //Now we need to fix positions and alt attribute of the images we've currently added
        $product = $this->productFactory->create();
        $product->setStoreId(0)->load($product_id);

        $this->_logger->info("INSERT IMAGES: Saving positions and alts attributes on photos of SKU {$sku}");
        //$productRepository = $objectManager->create('Magento\Catalog\Api\ProductRepositoryInterface');
        $existingMediaGalleryEntries = $product->getMediaGalleryEntries();
        if (count($existingMediaGalleryEntries)) {
            foreach ($existingMediaGalleryEntries as $key => $entry) {
                $current_image_name = $entry->getFile();
                foreach ($position_to_filename_array as $position => $position_to_filename_data) {
                    // If this image is one of those we've currently added then fix position and alt of this one!

                    if (strpos($current_image_name, $position_to_filename_data['image_name_no_ext']) !== false) {
                        //$existingMediaGalleryEntries[$key]['position'] = (string)$position;
                        //$existingMediaGalleryEntries[$key]['label'] = (string)$position_to_filename_data['image_alt'];
                        $this->updateCatalogProductImagePositionLabel($entry->getId(), $position, $position_to_filename_data['image_alt']);
                    }
                }
            }
            //$product->setMediaGalleryEntries($existingMediaGalleryEntries);
            //$productRepository->save($product);
            //$product = $objectManager->create('Magento\Catalog\Model\Product')->setStoreId(0)->load($product_id);
        }
        $this->_logger->info("INSERT IMAGES: Positions and alts attributes saved for SKU {$sku}");

        foreach ($collection as $image_data) {
            $image_data->setData('image_process_time',date("Y-m-d H:i:s"));
            $image_data->setData('meta_processed','1');
            $image_data->setData('image_status','processed');
            $image_data->save();
        }
        $this->_logger->info("INSERT IMAGES: Finished images import for SKU {$sku}");

        return true;
    }



    public function receiveFile($file){

        if (!$this->_connector)
            throw new \Exception('no connector joined to channel');

        $meta_file_id = $file->getId();

        $local_dir = $this->_connector->getLocalDir()."/".$meta_file_id;

        $this->_connector->setLocalDir($local_dir);

        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\Imagesflow');

        $collection = $model->getCollection();
        $collection->addFieldToFilter("meta_processed",0);
        $collection->addFieldToFilter("meta_file",$meta_file_id);
        $collection->addFieldToFilter("image_status","found");
        $collection->addFieldToFilter("image_queue_time",array('notnull' => true));
        $collection->load();

        $result = true;

        foreach ($collection as $image) {
            try{

                // Is unique filename enabled?
                $unique_filename_pattern = null;
                if ($this->imagesHelper->isUniqueUploadImagesActive()){
                    $unique_filename_pattern = "_unq".$image->getId();
                }

                $received = $this->_connector->receiveResource($image->getData('image_name'),null,FTP_BINARY,true,$unique_filename_pattern);
                if ($received) {
                    $image->setData('image_receive_time',date("Y-m-d H:i:s"));
                    $image->setData('image_status','received');
                    if ($received != $image->getData('image_name')) {
                        $log = $image->getData('log');
                        $log = trim("Image name fixed from '".$image->getData('image_name')."' to '{$received}'".$log."\n");
                        $image->setData('log',$log);
                        $image->setData('image_name',$received);
                    }
                    $image->save();
                } else {
                    $image->setData('image_status','found');
                    $log = $image->getData('log');
                    $log = $log."\n"."Image not received";
                    $image->setData('log',$log);
                    $image->save();
                    $result = false;
                }
            }catch (\Exception $e){
                $log = $image->getData('log');
                $log = $log."\n".$e->getMessage();
                $image->setData('log',$log);
                $image->save();
                $result = false;
                //$this->_logger->log('FTP',$e->getMessage());
            }
        }

        //Can't close the connection or the next cycle will fail
        //$this->_connector->close();

        return $result;

    }


    public function parse($file){
        if (!$this->_connector)
            throw new \Exception('no connector joined to channel');

        $meta_file_id = $file->getId();

        //$local_dir = $this->_connector->getLocalDir().DIRECTORY_SEPARATOR.$meta_file_id;
        //$this->_connector->setLocalDir($local_dir);

        $channel_config = $this->_connector->getConfig();
        $local_dir = $channel_config['local_dir'].DIRECTORY_SEPARATOR.$meta_file_id;

        //$this->directoryList->getPath('var');

        $csv_filename = $file->getName();
        $channel_name = $file->getChannel();
        $this->loadImageChannelConfig($channel_name);
        $this->filename_separator = $this->getImageSeparator();

        if (!$this->channel_photo_config) {
            throw new \Exception("Wrong images configuration");
        }

        $this->loadImagesListToParse($meta_file_id);

        foreach ($this->collection_of_images_to_parse as $data){
            $image_filename = $data->getImageName();
            $image_filename_no_ext = $this->removeExtensionFromFilename($image_filename);
            $loaded_data = $this->loadDataBasedOnFilename($image_filename,$channel_name);

            if (count($loaded_data) > 0) {
                $data->setData('image_sku',$loaded_data['sku']);

                //Handling not numeric "L1", "L2", "L3", "L4" special positions (pattern inside filename)
                //Giving custom high order (order is mandatory, numeric and unique)
                switch (strtolower($loaded_data['position'])) {
                    case "l1":
                        $loaded_data['position'] = 999;
                        break;
                    case "l2":
                        $loaded_data['position'] = 998;
                        break;
                    case "l3":
                        $loaded_data['position'] = 997;
                        break;
                    case "l4":
                        $loaded_data['position'] = 996;
                        break;
                    default:
                        #nothing to do
                }

                $data->setData('image_position',(int)$loaded_data['position']);
                $data->setData('image_alt',$loaded_data['alt']);
                $data->setData('image_type',implode("|",$loaded_data['image_type']));
                $data->setData('image_hide',$loaded_data['image_hide']);
                $data->setData('image_path',$local_dir);
                $data->setData('image_status','parsed');
                $data->save();
            } else {
                $data->setData('image_status','error');
                $data->save();
                $log = $file->getData('log');
                $log = $log."\n"."Can't parse image filename, wrong format! {$csv_filename}";
                $file->setData('log',$log);
                $file->save();
                continue;
            }
        }

        return true;
    }

    private function loadImageChannelConfig($channel_name){
        $this->channel_photo_config = null;
        if ($this->_parser->loadMap($channel_name)) {
            $config_mapping = $this->_parser->map->getFullMappedData();
            $this->channel_photo_config = $config_mapping;
        }
    }

    private function loadImagesListToParse($file_id){
        $model = \Magento\Framework\App\ObjectManager::getInstance()->create('FiloBlu\Flow\Model\Imagesflow');

        $collection = $model->getCollection();
        $collection->addFieldToFilter("meta_processed",0);
        $collection->addFieldToFilter("meta_file",$file_id);
        $collection->addFieldToFilter("image_status","received");
        $collection->load();

        $this->collection_of_images_to_parse = $collection;

    }

    private function loadDataBasedOnFilename($image_filename,$channel_name = null){
        $original_filename = $image_filename;

        // Remove extension
        $image_filename = $this->removeExtensionFromFilename($image_filename);

        $filename_exploded = explode($this->filename_separator, $image_filename);

        $sku_position = $this->getSkuFilenamePosition();
        $sku = "_unknown_";
        if (isset($filename_exploded[$sku_position])){
            $sku = $filename_exploded[$sku_position];
        }

        $image_position_position = $this->GetImagePositionPosition();

        $position = isset($filename_exploded[$image_position_position]) ? $filename_exploded[$image_position_position] : 1;

        $image_alt_positions = $this->GetImageAltPositions();
        $alt = array();
        if (count($image_alt_positions) > 0) {
            foreach ($image_alt_positions as $pos) {
                $alt[] = $filename_exploded[$pos];
            }
            $alt = implode($this->image_alt_separator, $alt);
        } else {
            $alt = "";
        }

        $image_type = $this->getImageType($position);
        $image_hide = $this->getImageHide($position);

        $this->array_of_data_for_csv[$sku][$position]['filename'] = $original_filename;
        $this->array_of_data_for_csv[$sku][$position]['path'] = $this->_connector->getLocalDir();
        $this->array_of_data_for_csv[$sku][$position]['position'] = $position;
        $this->array_of_data_for_csv[$sku][$position]['alt'] = $alt;
        $this->array_of_data_for_csv[$sku][$position]['image_type'] = $image_type;
        $this->array_of_data_for_csv[$sku][$position]['image_hide'] = $image_hide;
        $this->array_of_data_for_csv[$sku][$position]['sku'] = $sku;

        // Start dispatch event to handle change of this data if needed
        $event_dispatch_data = new \Magento\Framework\DataObject();
        $event_dispatch_data->setData('channel_name',$channel_name);
        $event_dispatch_data->setData('image_filename',$original_filename);
        $event_dispatch_data->setData('image_filename_noext',$image_filename);
        $event_dispatch_data->setData('filename_separator',$this->filename_separator);
        $event_dispatch_data->setData('loaded_data',$this->array_of_data_for_csv[$sku][$position]);
        $this->eventManager->dispatch('flow_images_parse_load_data_based_on_filename_after',['event_dispatch_data'=>$event_dispatch_data]);
        // If the value is changed then it will be updated here
        $this->array_of_data_for_csv[$sku][$position] = $event_dispatch_data->getData('loaded_data');
        // End dispatch event, the data may have been changed here

        return $this->array_of_data_for_csv[$sku][$position];
    }


    private function getSkuFilenamePosition(){
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] == 'sku') {
                return $key;
            }
        }

        //Handle error missing SKU on config here, block all flow
    }

    private function GetImagePositionPosition(){
        $config = $this->channel_photo_config;
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] == 'position') {
                return $key;
            }
        }

        //Handle error missing SKU on config here, block all flow
    }

    private function GetImageAltPositions(){
        $config = $this->channel_photo_config;
        $positions = array();
        foreach ($config['fields'] as $key => $value) {
            if ($value['type'] == 'alt') {
                $positions[] = $key;
            }
        }

        return $positions;

        //Handle error missing SKU on config here, block all flow
    }

    private function getImageType($position){
        $config = $this->channel_photo_config;

        if (isset($config['types'][$position]['type'])) {
            return $config['types'][$position]['type'];
        }

        return array();

    }

    private function getImageHide($position){
        $config = $this->channel_photo_config;

        if (isset($config['types'][$position]['hide']) && $config['types'][$position]['hide'] == "true") {
            return 1;
        } else {
            return 0;
        }

    }

    private function getImageSeparator(){
        $config = $this->channel_photo_config;

        $has_separator = (isset($config['options']['separator']) && trim($config['options']['separator']) != '');
        return ($has_separator) ? $config['options']['separator'] :"_";

    }

    private function removeCatalogProductImageByPath($path){

        $file = BP . "/pub/media/catalog/product{$path}";
        $pathinfo = pathinfo($file);
        $filename = $pathinfo['basename'];

        $resource = \Magento\Framework\App\ObjectManager::getInstance()->create('\Magento\Framework\App\ResourceConnection');
        $connection = $resource->getConnection(\Magento\Framework\App\ResourceConnection::DEFAULT_CONNECTION);

        $query = "DELETE FROM catalog_product_entity_media_gallery WHERE value = '{$path}'";
        $result = $connection->query($query);

        if (!$result) {
            $this->_logger->info("INSERT IMAGES: NOT Removed from catalog_product_entity_media_gallery {$path}");
            return false;
        } else {
            $this->_logger->info("INSERT IMAGES: Removed from catalog_product_entity_media_gallery {$path}");
        }

        $ProductMediaImageAttributeIds = $this->getProductMediaImageAttributeIds();
        if($ProductMediaImageAttributeIds){
            $add_where = " AND attribute_id IN ({$ProductMediaImageAttributeIds}) ";
        }

        // Removing also the "photo tags" (base, small, thumbnail) or they will be left there
        $query = "DELETE FROM catalog_product_entity_varchar WHERE value = '{$path}' {$add_where}";
        $result = $connection->query($query);

        if (!$result) {
            $this->_logger->info("INSERT IMAGES: NOT Removed from catalog_product_entity_varchar {$path}");
            return false;
        } else {
            $this->_logger->info("INSERT IMAGES: Removed TAG from catalog_product_entity_varchar {$path}");
        }

        if (is_file($file)) {
            unlink($file);
            $this->_logger->info("INSERT IMAGES: Removed file from disk {$file}");
        } else {
            $this->_logger->info("INSERT IMAGES: Removing file from disk {$file} not found, skipping");
        }

        $this->removeImageCacheFromImageFileName($path);

        return true;
    }

    /*
     * $filename is column 'value' from 'catalog_product_entity_media_gallery'
     */
    private function removeImageCacheFromImageFileName($filename = null){

        if(!$this->imagesHelper->isClearImageCacheActive() || !$filename){
            return true;
        }

        // Array of all cache main folders
        try {
            $main_cache_image_paths = $this->getAllCacheImageMainPaths();
        } catch(\Exception $e) {
            $this->_logger->info("ERROR IMAGES: removeImageCacheFromImageFileName " . $e->getMessage());
            return false;
        }

        if(!isset($main_cache_image_paths) || !$main_cache_image_paths || count($main_cache_image_paths) == 0){
            return false;
        }

        foreach ($main_cache_image_paths as $path) {

            // Safe check
            $path = rtrim($path, DIRECTORY_SEPARATOR);

            $full_path = $this->cache_folder . $path . $filename;

            if(file_exists($full_path)){
                // Try to delete the cache
                unlink($full_path);
                $this->_logger->info("INSERT IMAGES: Removed cache image {$full_path}");
            }
        }

    }


    public function getAllCacheImageMainPaths(){
        // Give me an array of all the folders inside the cache folder
        // Going deep 3
        $image_cache_paths_array = $this->imagesHelper->foldersMap($this->cache_folder,3);

        // Transform the array paths into string path checking also the md5 of the last folder
        $array_of_builded_paths = array();
        $this->imagesHelper->collapse("",$image_cache_paths_array,$array_of_builded_paths);

        return $array_of_builded_paths;
    }

    private function updateCatalogProductImagePositionLabel($db_row_id, $position, $label){

        $resource = \Magento\Framework\App\ObjectManager::getInstance()->create('\Magento\Framework\App\ResourceConnection');
        $connection = $resource->getConnection(\Magento\Framework\App\ResourceConnection::DEFAULT_CONNECTION);

        $query = "UPDATE catalog_product_entity_media_gallery_value SET position = {$position}, label = '{$label}' WHERE value_id = {$db_row_id}";
        $result = $connection->query($query);

        if (!$result) {
            return false;
        }

        return true;
    }

    public function removeExtensionFromFilename($string){
        $p = strrpos($string, '.');
        if ($p !== false){
            $string = substr($string, 0, $p);
        }

        return $string;
    }

    /**
     * @return string of ids comma separated
     */
    private function getProductMediaImageAttributeIds(){
        /** @var  $coll \Magento\Eav\Model\ResourceModel\Entity\Attribute\Collection */
        $coll = \Magento\Framework\App\ObjectManager::getInstance()->create(\Magento\Eav\Model\ResourceModel\Entity\Attribute\Collection::class);
        $coll->addFieldToFilter('frontend_input', 'media_image');
        $attrAll = $coll->load()->getItems();
        $result = array();
        foreach($attrAll as $attr){
            $result[] = $attr->getAttributeId();
        }

        $return = false;
        if(count($result)){
            $return = implode(",",$result);
        }

        return $return;
    }



}
